# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import numpy as np
from domainbed.lib import misc


def _define_hparam(hparams, hparam_name, default_val, random_val_fn):
    hparams[hparam_name] = (hparams, hparam_name, default_val, random_val_fn)


def _hparams(algorithm, dataset, random_seed):
    """
    Global registry of hyperparams. Each entry is a (default, random) tuple.
    New algorithms / networks / etc. should add entries here.
    """
    SMALL_IMAGES = ['Debug28', 'RotatedMNIST', "MNIST_Acause",
    "MNIST_Aind",
    "MNIST_AcauseUAind"]

    hparams = {}

    def _hparam(name, default_val, random_val_fn):
        """Define a hyperparameter. random_val_fn takes a RandomState and
        returns a random hyperparameter value."""
        assert(name not in hparams)
        random_state = np.random.RandomState(
            misc.seed_hash(random_seed, name)
        )
        hparams[name] = (default_val, random_val_fn(random_state))

    # Unconditional hparam definitions.

    _hparam('data_augmentation', True, lambda r: True)
    if dataset == "SmallNORB":
        _hparam('resnet18', True, lambda r: True)
    else:
        _hparam('resnet18', False, lambda r: False)
    _hparam('resnet_dropout', 0., lambda r: r.choice([0., 0.1, 0.5]))
    _hparam('class_balanced', False, lambda r: False)
    # TODO: nonlinear classifiers disabled
    _hparam('nonlinear_classifier', False,
            lambda r: bool(r.choice([False, False])))

    # Algorithm-specific hparam definitions. Each block of code below
    # corresponds to exactly one algorithm.

    if algorithm in ['DANN', 'CDANN']:
        _hparam('lambda', 1.0, lambda r: r.choice([0.01, 0.1, 1., 10., 100.]))
        _hparam('weight_decay_d', 0., lambda r: 10**r.uniform(-6, -2))
        _hparam('d_steps_per_g_step', 1, lambda r: r.choice([1., 2., 4., 8.]))
        _hparam('grad_penalty', 0., lambda r: r.choice([0.01, 0.1, 1., 10.]))
        _hparam('beta1', 0.5, lambda r: r.choice([0., 0.5]))

    elif algorithm == 'Fish':
        _hparam('meta_lr', 0.5, lambda r:r.choice([0.05, 0.1, 0.5]))

    elif algorithm == "RSC":
        _hparam('rsc_f_drop_factor', 1/3, lambda r: r.uniform(0, 0.5))
        _hparam('rsc_b_drop_factor', 1/3, lambda r: r.uniform(0, 0.5))

    elif algorithm == "SagNet":
        _hparam('sag_w_adv', 0.1, lambda r: r.choice([0.01, 0.1, 1.0, 10.0]))

    elif algorithm == "IRM":
        _hparam('irm_lambda', 1e2, lambda r: r.choice([0.01, 0.1, 1., 10., 100.])) 
        if dataset == 'CelebA_Blonde':
            _hparam('irm_penalty_anneal_iters', 500,
                    lambda r: int(10**r.uniform(0, 3.5)))
        else:
            _hparam('irm_penalty_anneal_iters', 100,
                    lambda r: int(10**r.choice([1, 2, 3]))) 
    
    elif algorithm == "IB_ERM":
        _hparam('ib_lambda', 1e2, lambda r: r.choice([0.01, 0.1, 1., 10., 100.]))
        _hparam('ib_penalty_anneal_iters', 500,
                lambda r: int(10**r.choice([1, 2, 3])))

    elif algorithm == "IB_IRM":
        _hparam('irm_lambda', 1e2, lambda r: r.choice([0.01, 0.1, 1., 10., 100.]))
        _hparam('irm_penalty_anneal_iters', 500,
                lambda r: int(10**r.choice([1, 2, 3])))
        _hparam('ib_lambda', 1e2, lambda r: r.choice([0.01, 0.1, 1., 10., 100.]))
        _hparam('ib_penalty_anneal_iters', 500,
                lambda r: int(10**r.choice([1, 2, 3])))

    elif algorithm == "Mixup":
        _hparam('mixup_alpha', 0.2, lambda r: r.choice([0.1, 1.0, 10.0]))

    elif algorithm == "GroupDRO":
        _hparam('groupdro_eta', 1e-2, lambda r: r.choice([0.001, 0.01, 0.1]))

    elif algorithm == "MMD" or algorithm == "CORAL" or algorithm == "CACM_ACause" or algorithm == "CACM_Aind" or or algorithm == "CACM_AcauseUAind":
        _hparam('mmd_gamma', 100., lambda r: r.choice([0.1, 1., 10., 100.])) 
        _hparam('sigma', .01, lambda r: r.choice([0.01, 0.0001, 0.000001])) 
        _hparam('mmd_gamma_ind', 10., lambda r: r.choice([0.1, 1., 10., 100.])) 
        
    
    elif algorithm == "MLDG":
        _hparam('mldg_beta', 1., lambda r: r.choice([0.1, 1.0, 10.0]))

    elif algorithm == "MTL":
        _hparam('mtl_ema', .99, lambda r: r.choice([0.5, 0.9, 0.99, 1.]))

    elif algorithm == "VREx":
        _hparam('vrex_lambda', 1e2, lambda r: r.choice([0.01, 0.1, 1., 10., 100.])) # modify
        if dataset == 'CelebA_Blonde':
            _hparam('vrex_penalty_anneal_iters', 500,
                    lambda r: int(10**r.uniform(0, 3.5)))
        else:
            _hparam('vrex_penalty_anneal_iters', 10,
                    lambda r: int(10**r.choice([1, 2, 3]))) # modify

    # Dataset-and-algorithm-specific hparam definitions. Each block of code
    # below corresponds to exactly one hparam. Avoid nested conditionals.

    if dataset in SMALL_IMAGES:
        _hparam('lr', 1e-5, lambda r: r.choice([1e-2, 1e-3, 1e-4, 1e-5])) 
    else:
        _hparam('lr', 1e-5, lambda r: r.choice([1e-2, 1e-3, 1e-4, 1e-5])) 

    if dataset in SMALL_IMAGES:
        _hparam('weight_decay', 0., lambda r: 0.)
    else:
        _hparam('weight_decay', 0., lambda r: 10**r.uniform(-6, -2))

    if dataset in SMALL_IMAGES:
        _hparam('batch_size', 64, lambda r: 64) 
    elif dataset == 'Waterbirds_Acause' or dataset == 'Waterbirds_Multiattr':
        _hparam('batch_size', 16, lambda r: 16)
    elif dataset == 'Waterbirds_Acause_CACM' or dataset == 'Waterbirds_Multiattr_CACM':
        _hparam('batch_size', 32, lambda r: 32)
    else:
        _hparam('batch_size', 128, lambda r: 128) 

    if algorithm in ['DANN', 'CDANN'] and dataset in SMALL_IMAGES:
        _hparam('lr_g', 1e-3, lambda r: 10**r.uniform(-4.5, -2.5))
    elif algorithm in ['DANN', 'CDANN']:
        _hparam('lr_g', 5e-5, lambda r: 10**r.uniform(-5, -3.5))

    if algorithm in ['DANN', 'CDANN'] and dataset in SMALL_IMAGES:
        _hparam('lr_d', 1e-3, lambda r: 10**r.uniform(-4.5, -2.5))
    elif algorithm in ['DANN', 'CDANN']:
        _hparam('lr_d', 5e-5, lambda r: 10**r.uniform(-5, -3.5))

    if algorithm in ['DANN', 'CDANN'] and dataset in SMALL_IMAGES:
        _hparam('weight_decay_g', 0., lambda r: 0.)
    elif algorithm in ['DANN', 'CDANN']:
        _hparam('weight_decay_g', 0., lambda r: 10**r.uniform(-6, -2))

    return hparams


def default_hparams(algorithm, dataset):
    return {a: b for a, (b, c) in _hparams(algorithm, dataset, 0).items()}


def random_hparams(algorithm, dataset, seed):
    return {a: c for a, (b, c) in _hparams(algorithm, dataset, seed).items()}
